/*
	morse - output characters as morse code using the PC speaker

	NOTES
	statistics on (actual delay - requested delay) for usleep():
	mean extra delay is 14922 usec
	earliest = 7075 usec
	latest = 52726 usec

	$Id: morse.c,v 1.5 1997/12/06 01:13:50 jrv Exp jrv $
	
	$Log: morse.c,v $
	Revision 1.5  1997/12/06 01:13:50  jrv
	  DEL has code ........
	  Raising audio frequency by a semitone for upper case letters.
	  Lowering frequency by a tone for special characters.
	  -s or --string switch to supply string on command line.

	Revision 1.4  1997/12/04 01:52:29  jrv
	  Getting input with read() instead of fgets().
	  Checking input before each character, and flushing if more input is
	available.

	Revision 1.3  1997/12/04 01:35:34  jrv
	  Using getopt_long.
	  --rate and --frequency set the transmission rate and audio frequency.
	  --help and --version do the usual things.

	Revision 1.2  1997/11/24 02:33:32  jrv
	  Moved measurement delay code to calculate_bit_times().
	  -d option to report the measured delays.
	  checkshiftstate() checks are #ifdef'd out by default.
	  Special characters are spelled out (at a lower pitch).

 * Revision 1.1  1995/04/20  01:18:52  jrv
 * Initial revision
 *
*/

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/kd.h>

#include "morse.h"
#ifdef FEEDBACK
#include "pty.h"
#endif
#include "debug.h"
ASSERTFILE;

#define BUFFERSIZE 10000

#define KSHIFT 1
#define KCNTL 4
#define KALT 8

#define ITER 5

static enum {SPEAKING=9321, SKIMMING, SILENT} speakMode=SPEAKING;

int getshiftstate(void);
int checkshiftstate(void);

#define NONCE_ERROR "" /* "..--" */

static char *code[]={
/* SP */ NONCE_ERROR,
/* ! */ "exclamation ",
/* " */ ".-..-.",
/* # */ "pound ",
/* $ */ "dollar ",
/* % */ "percent ",
/* & */ "ampersand ",
/* ' */ "apostrophe ",
/* ( */ "left parenthesis ",
/* ) */ "right parenthesis ",
/* * */ "asterisk ",
/* + */ "plus ",
/* , */ "--..--",
/* - */ "-...-",
/* . */ ".-.-.-",
/* / */ "-..-.",

/* 0-9 */ 
"-----", ".----", "..---", "...--", "....-", ".....", "-....",
"--...", "---..", "----.",
/* : */ "---...",
/* ; */ "-.-.-.",
/* < */ "less than ",
/* = */ "equal ",
/* > */ "greater than ",
/* ? */ "..--..",
/* @ */ "at ",
/* A-Z */
".-", "-...", "-.-.", "-..", ".", "..-.", "--.", "....", "..", ".---",
"-.-", ".-..", "--", "-.", "---", ".--.", "--.-", ".-.", "...", "-",
"..-", "...-", ".--", "-..-", "-.--", "--..",
/* [ */ "left bracket ",
/* \ */ "back slash ",
/* ] */ "right bracket ",
/* ^ */ "caret ",
/* _ */ "underscore ",
/* ` */ "grave accent ",
/* a-z */
".-", "-...", "-.-.", "-..", ".", "..-.", "--.", "....", "..", ".---",
"-.-", ".-..", "--", "-.", "---", ".--.", "--.-", ".-.", "...", "-",
"..-", "...-", ".--", "-..-", "-.--", "--..",
/* { */ "left brace ",
/* | */ "pipe ",
/* } */ "right brace ",
/* ~ */ "tilde ",
/* DEL */ "........",
};

int earliest=1000000, latest=-1000000, extra, overhead=0;
static int report_delay = 0;
int morse_initialized = 0;
double morse_wpm = 12.;
int highlight_upper_case = 1;
double morse_frequency = 440*2;
char *string=NULL;
static int pad_usec = 15000;
static int dit_usec;
static int dah_usec; 
static int interchar_usec;
static int interword_usec;
static vt_active = 1;	/* nonzero while controlling 
				virtual console is active */
double note[12] = {261.63, 277.18, 293.66, 311.13, 329.63, 349.23, 
		370, 392, 415.30, 440, 466.16, 493.88};

double note_frequency(int k)
{
	int power = 0;
	while(k < 0) {power--; k += 12;}
	while(k > 11) { power++; k -= 12;}
	if(power >= 0) return note[k]*(1<<power);
	return note[k]/(1<<(-power));
}

void calculate_bit_times(void)
{
	double long_dit_usec;

	if(morse_wpm >= MIN_CHAR_WPM) {
		dit_usec = 60000000/(50*morse_wpm);
		dah_usec = 
		interchar_usec = 3*dit_usec;
		interword_usec = 7*dit_usec;
	} else {
	  /* for low rates, we keep the same timing within a character
             but stretch the intercharacter and interword pauses */
		dit_usec = 60000000/(50*MIN_CHAR_WPM);
		dah_usec = 3*dit_usec;
		long_dit_usec = (60000000/morse_wpm - 31*dit_usec)/19;
		interchar_usec = 3*long_dit_usec;
		interword_usec = 7*long_dit_usec;
	}

#ifdef MEASURE_DELAYS
	{
	  struct timeval prestart, start, finish;
	  int requested[1000], actual[1000], i, time_usec, dif;
	

	  for (i = 0; i < ITER; i++)
	    {
	      /* requested times are uniformly distributed between 0
                 and 60 msec */
	      requested[i] =
		time_usec = (int)((double)random()/RAND_MAX*60000);
	      gettimeofday(&prestart, NULL);
	      gettimeofday(&start, NULL);
	      usleep(time_usec);
	      gettimeofday(&finish, NULL);
	      overhead += (start.tv_sec - prestart.tv_sec)*1000000 + 
		start.tv_usec - prestart.tv_usec;
	      actual[i] = (finish.tv_sec - start.tv_sec)*1000000 + 
		finish.tv_usec - start.tv_usec;
	      dif = actual[i] - requested[i];
	      extra += dif;
	      if(dif < earliest) earliest = dif;
	      if(dif > latest) latest = dif;
	    }
	  pad_usec = extra/ITER;
	}
#endif /* MEASURE_DELAYS */
	dit_usec -= pad_usec;
	dah_usec -= pad_usec;
	interchar_usec -= pad_usec;
	interword_usec -= pad_usec;
	morse_initialized = 1;
}

void xusleep(int duration)
{
	if(vt_active) usleep(duration);
}

/*
	emit tone of "duration" usec, at "frequency" Hz, returning when 
	tone is finished.  (If /dev/tty0 is no longer the active VT,
	return immediately.) 
*/
void tone(int duration, double frequency)
{
	int soundcode;
	int fd;
	static reported;
	ASSERT(1193180/frequency < 65536);
	ASSERT(1193180/frequency > 1);
	ASSERT(duration >= 0);
	ASSERT(duration < 65536000);

	fd = open("/dev/tty0", O_RDONLY);
	if(!reported)
		{FILE *dfile=fopen("debug","a");
		fprintf(dfile,"fd = %d is %s\n",
			fd, isatty(fd) ? ttyname(fd) : "not a tty");
		fclose(dfile);
		reported = 1;
		}
	if(fd<0) {perror("tone: opening console"); exit(1);}
	soundcode = ((duration/1000)<<16) + (int)(1193180/frequency);
	if(ioctl(fd, KDMKTONE, soundcode))
		{if(errno == EPERM) 
			{vt_active = 0;
			goto tone_finished;
			}
		perror("tone: ioctl(console, KDMKTONE, ...)");
		exit(1);
		}
	vt_active = 1;
	usleep(duration);
tone_finished:
	close(fd);
}

/*
	output string pointed to by "s" in Morse code  
*/
void write_morse(char const *s)
{	char *t;
	int c, duration;
	double send_frequency;
	fd_set rfd, wfd, efd;
	static struct timeval timeout={0,0};

	if(!morse_initialized) calculate_bit_times();
	while((c = *s++) != 0)
		{
		  FD_ZERO(&rfd);
		  FD_SET(0, &rfd);
		  FD_ZERO(&wfd);
		  FD_ZERO(&efd);
		  if (select(1, &rfd, &wfd, &efd, &timeout))
		      break;	/* more input is available */
		  
#ifdef FEEDBACK
		switch(checkshiftstate())
			{case SKIMMING:
				if(c == '\n') 
					{tone(dit_usec, morse_frequency/2);
					xusleep(dit_usec);
					}
			case SILENT:
				continue;
			case SPEAKING:
				break;
			}
#endif /* FEEDBACK */
		if(c < ' ') continue;
		else if (c == ' ') xusleep(interword_usec);
		else if (' ' < c && c < 128)
			{send_frequency = morse_frequency;
			if (highlight_upper_case && c >= 'A' && c <= 'Z')
			  send_frequency *= 1.05946309435930;
			t =  code[c - ' '];
			if(*t == '\0') 
				{tone(dah_usec, morse_frequency/3); 
				continue;
				}
			if (*t == '.' || *t == '-')
			  while(1)
			    {
			      if(*t == '.') duration = dit_usec;
			      else duration = dah_usec;
			      tone(duration, send_frequency);
			      if(*++t == '\0') break;
			      xusleep(dit_usec);
			    }
			else 
			  {
			    /* spell out a special character, at a
                               lower pitch */
			    morse_frequency /= 1.12246204830937;
			    write_morse(t);
			    morse_frequency *= 1.12246204830937;
			  }
			xusleep(interchar_usec);
			}
		}
}

#ifdef FEEDBACK
/*
	fetch and return shift state

	weights:
		shift	1
		l-cntl	4
		l-alt	8
*/
int getshiftstate(void)
{	int shift=6;
	extern fdConsole;
	static reported;
	if(!reported)
		{FILE *dfile=fopen("debug","a");
		fprintf(dfile,"fdConsole = %d is %s\n",
			fdConsole, isatty(fdConsole) ? ttyname(fdConsole) : "not a tty");
		fclose(dfile);
		reported = 1;
		}

	if(ioctl(fdConsole, TIOCLINUX, &shift) < 0)
		err_sys("shift_state: ioctl");

	return shift;
}

int checkshiftstate(void)
{	static int prevShift;
	int curShift = getshiftstate();
	switch(speakMode)
		{case SPEAKING:
			if(!(prevShift & KALT) & (curShift & KALT)) 
				speakMode = SKIMMING;
			break;
		case SKIMMING:
			if(!(prevShift & KALT) & (curShift & KALT)) 
				speakMode = SPEAKING;
			break;
		case SILENT:
			/*******  what to do?  ********/
			break;
		}
	prevShift = curShift;
	return speakMode;
}
#endif /* FEEDBACK */


#ifdef TEST


int main(int argc, char **argv)
{
  enum {C, Csharp, D, Dsharp, E, F, Fsharp, G, Gsharp, A, Asharp, B} n;
  char *t, buf[BUFFERSIZE+1];
  int c, num;
  static char revision_string[]="$Revision: 1.5 $";
  char *revision = revision_string, *s;

  enum {HELP=257};
  int longind;

  static struct option longopts[] =
    {
      {"debug", 0, 0, 'd'},
      {"rate", 1, 0, 'r'},
      {"frequency", 1, 0, 'f'},
      {"string", 1, 0, 's'},
      {"uppercase", 0, 0, 'u'},
      {"version", 0, 0, 'v'},
      {"help", 0, 0, HELP},
      {0, 0, 0, 0}
    };

				/* isolate the version number */
  revision += strlen("$Revision: ");
  if ((s = strchr(revision, ' '))) *s = 0;

  while((c = getopt_long(argc, argv, "df:hr:s:v", longopts, &longind)) != EOF)

    {
      switch(c)
	{
	case 'd':
	  report_delay = 1;
	  break;
	case 'f':
	  morse_frequency = atof(optarg);
	  break;
	case 'r':
	  morse_wpm = atof(optarg);
	  break;
	case 's':
	  string = optarg;
	  break;
	case 'u':
	  highlight_upper_case = 1;
	  break;
	case 'v':
	  printf("morse %s\n", revision);
	  exit(0);
	case HELP:
	  puts(
"morse - emit morse code using PC speaker\n"
"usage:  morse  [options]  [<textfile]\n"
"options:\n"
"  -d, --debug          enable debugging output\n"
"  -f, --frequency F    set base audio frequency (Hz)\n"
"  -r, --rate R         set transmission rate (wpm)\n"
"  -v, --version        output morse program version and exit\n"
"  --help               output this help message and exit\n");
	  exit(0);
	default:
	  fprintf(stderr, "morse - emit morse code\n"
		  "for usage, try   morse --help\n");
	  exit(1);
	}
    }
  if (report_delay)
    {
      calculate_bit_times();
      printf("mean timing overhead: %d usec\n", overhead/ITER);
      printf("    mean extra delay: %d usec\n", extra/ITER);
      printf("               least: %d usec\n", earliest);
      printf("                most: %d usec\n", latest);
      exit(0);
    }

  if (string) 
    {
      write_morse(string);
      exit(0);
    }

  while((num = read(0, buf, BUFFERSIZE)) > 0)
    {
      buf[num] = 0;
      write_morse(buf);
    }
  exit(0);

#ifdef NOTES
  for (n = C; n <= B; n++)
    tone(130000, note[n]*2);
  usleep(500000);
  tone(230000, note[C]);
  tone(230000, note[D]);
  tone(230000, note[E]);
  tone(230000, note[F]);
  tone(230000, note[G]);
  tone(230000, note[A]);
  tone(230000, note[B]);
  tone(230000, note[C]*2);
  return 0;
#endif
}

#endif /* TEST */

